#include <iostream>
#include <vector>
#include <algorithm>

using namespace std;

using point = pair<int, int>;

int sprod(point v, point w) {
	return v.first * w.first + v.second * w.second;
}

int vprod(point v, point w) {
	return v.first * w.second - v.second * w.first;
}

point vec(point a, point b) {
	return {b.first - a.first, b.second - a.second};
}

int norm2(point a) {
	return a.first * a.first + a.second * a.second;
}

// 1. orijentacija
int orient(point a, point b, point c) {
	return vprod(vec(a, b), vec(a, c));
}

bool colinear(point a, point b, point c) {
	return orient(a, b, c) == 0;
}

// 2. tacka u trouglu?
bool in_angle(point p, point a, point b, point c) {
	if(orient(a, b, c) < 0)
		swap(b, c);
	return orient(a, b, p) >= 0 && orient(a, p, c) >= 0;
}

bool in_tri(point p, point a, point b, point c) {
	return in_angle(p, a, b, c) && in_angle(p, b, c, a) && in_angle(p, c, a, b);
}

// 3. provera konveksnosti
bool is_convex(vector<point>& p) {
	int n = p.size();
	bool pos = false, neg = false;
	for(int i = 0; i < n; i++) {
		int j = (i + 1) % n;
		int k = (i + 2) % n;
		int ori = orient(p[i], p[j], p[k]);
		if(ori > 0)
			pos = true;
		if(ori < 0)
			neg = true;
	}
	return !(pos && neg);
}

// 4. konstrukcija prostog mnogougla (kad nema kolinearnih trojki tacaka)
void make_convex(vector<point>& p) {
	swap(p[0], *min_element(p.begin(), p.end()));
	sort(p.begin() + 1, p.end(), [&p](point a, point b) { return orient(p[0], a, b) > 0; });
}

// 5. Jarvis march O(nh)
vector<int> jarvis(vector<point>& p) {
	vector<int> hull;

	int t = min_element(p.begin(), p.end()) - p.begin();
	do {
		hull.push_back(t);
		int mint = (t + 1) % p.size();
		for(int i = 0; i < p.size(); i++)
			if(i != t && orient(p[t], p[mint], p[i]) > 0)
				mint = i;
		t = mint;
	} while(t != hull[0]);

	return hull;
}

// 6. Graham scan O(nlogn)
vector<int> graham(vector<point>& p) {
	swap(p[0], *min_element(p.begin(), p.end()));
	sort(p.begin() + 1, p.end(), [&p](point a, point b) { return orient(p[0], a, b) > 0; });

	vector<int> hull(1, 0);
	for(int i = 1; i < p.size(); i++) {
		while(hull.size() > 1 && orient(p[hull[hull.size() - 2]], p[hull[hull.size() - 1]], p[i]) < 0)
			hull.pop_back();
		hull.push_back(i);
	}

	return hull;
}

// 7. tacka u konveksnom?
bool in_convex(point t, vector<point>& p) {
	int i = lower_bound(p.begin() + 2, p.end(), t, [&p](point a, point b){ return orient(p[0], a, b) > 0; }) - p.begin();
	if(i == p.size())
		return false;
	return in_tri(t, p[0], p[i - 1], p[i]);
}

// 8. povrsina trougla (sa znakom)
float area(point a, point b, point c) {
	return (float) vprod(vec(a, b), vec(a, c)) / 2;
}

// 9. povrsina (prostog, moze i nekonveksnog) mnogougla
float area(vector<point> p) {
	float a = 0;
	for(int i = 2; i < p.size(); i++)
		a += area(p[0], p[i - 1], p[i]);
	return a;
}

int main() {
	point a = {0, 0};
	point b = {5, 0};
	point c = {0, 5};
	point p = {1, 1};
	point q = {3, 3};

	cout << in_tri(p, a, b, c) << '\n';
	cout << in_tri(a, a, b, c) << '\n';
	cout << in_tri(q, a, b, c) << '\n';
	cout << in_tri(p, a, c, b) << '\n';
	cout << in_tri(a, a, c, b) << '\n';
	cout << in_tri(q, a, c, b) << '\n';

	vector<point> conv = {{0, 0}, {2, 0}, {3, 1}, {2, 2}, {0, 2}};
	cout << is_convex(conv) << '\n';

	vector<point> nconv = {{0, 0}, {1, 1}, {2, 0}, {2, 3}, {0, 3}};
	cout << is_convex(nconv) << '\n';

	vector<point> pts = {{0, 0}, {1, 0}, {0, 1}, {1, 1}};
	make_convex(pts);
	for(auto x : pts)
		cout << x.first << ", " << x.second << '\n';
	cout << '\n';

	vector<point> pts2 = {{0, 0}, {2, 1}, {0, 4}, {4, 0}, {3, 1}, {4, 3}};
	vector<int> hull = jarvis(pts2);
	for(int i : hull)
		cout << pts2[i].first << ", " << pts2[i].second << '\n';
	cout << '\n';

	cout << in_convex({1, 2}, conv) << '\n';

	return 0;
}
